/*
 * Generic object event notification support.
 *
 * Copyright (C) 2021, 2022, 2023 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#pragma once

#include <condition_variable>
#include <list>
#include <mutex>
#include <set>
#include <vector>

#include <resource/resource.h>
#include <systypes/base.h>

#include "notifier_interface.h"



/* Collection types. */

typedef std::list<notifiable_t *> NotifiableObjectQueue;
typedef std::set<notifiable_t *> NotifiableObjects;



/* An object for monitoring object event notifications. */

class ObjectNotifier
{
protected:
    /* Locking for the affected queue and monitored object registry. */

    std::mutex _affected_lock, _monitored_lock;
    std::condition_variable _condition;

    /* Objects affected by notifications. */

    NotifiableObjectQueue _queued;
    NotifiableObjects _affected, _monitored;

public:
    virtual ~ObjectNotifier();

    /* Local operations. */

    virtual long subscribe(notifiable_t *object, notify_flags_t flags);

    virtual long unsubscribe(notifiable_t *object);

    virtual long wait(notifiable_t **object);

    virtual long wait_object(notifiable_t *object);

    /* Helper methods. */

    virtual void notify(notifiable_t *object);

    virtual void release(notifiable_t *object);
};



/* Collection types. */

typedef std::set<ObjectNotifier *> ObjectNotifiers;



/* Object-specific notifier resource. */

class NotifierResource : public Notifier, public Resource
{
protected:
    /* Locking for the resource. */

    std::mutex _lock;
    std::condition_variable _condition;

    /* Notifiers and the monitored object. */

    ObjectNotifiers _notifiers;
    notifiable_t *_object;

    /* Pending notification status. */

    bool _pending = false;

    /* Utility methods. */

    virtual void _notify();

    virtual void _release();

public:
    l4_cap_idx_t endpoint = L4_INVALID_CAP;

    explicit NotifierResource(notifiable_t *object)
    : _object(object)
    {
    }

    virtual ~NotifierResource();

    /* Resource methods. */

    virtual void close();

    /* Server details. */

    virtual ipc_server_default_config_type config();

    virtual void *interface()
    { return static_cast<Notifier *>(this); }

    /* Notifier methods. */

    virtual long notify(notify_flags_t flags, notify_values_t values);

    /* Local operations. */

    virtual long add(ObjectNotifier *notifier, notify_flags_t flags);

    virtual long remove(ObjectNotifier *notifier);

    virtual long wait();
};



/* Helper functions. */

ObjectNotifier *notifier_get_task_notifier();

ObjectNotifier *notifier_get_local_notifier();

// vim: tabstop=4 expandtab shiftwidth=4
