/*
 * Generic object event notification support.
 *
 * Copyright (C) 2021, 2022, 2023 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <ipc/cap_alloc.h>
#include <ipc/server.h>
#include <resource/resource_server.h>

#include "notification_client.h"
#include "notifier.h"
#include "notifier_server.h"



/* Lock protecting per-task notifier access. */

static std::mutex _task_lock;

/* Per-task storage for specific waiting operations. */

static ObjectNotifier *_notifier = NULL;



/* Return the per-task notifier for object-specific waiting operations. */

ObjectNotifier *notifier_get_task_notifier()
{
  std::lock_guard<std::mutex> guard(_task_lock);

  /* Start any new notifier. */

  if (_notifier == NULL)
    _notifier = new ObjectNotifier;

  return _notifier;
}

/* Return a local notifier for general object waiting operations. */

ObjectNotifier *notifier_get_local_notifier()
{
  return new ObjectNotifier;
}



/* Virtual destructor required for base class instance reference deletion. */

ObjectNotifier::~ObjectNotifier()
{
  /* Remove this notifier from the individual notifier resources. */

  NotifiableObjects::iterator it;

  for (it = _monitored.begin(); it != _monitored.end(); it++)
  {
    notifiable_t *object = *it;

    if (object->handler != NULL)
    {
      NotifierResource *resource = reinterpret_cast<NotifierResource *>(object->handler);
      resource->remove(this);
    }
  }

  _monitored.clear();

  /* Handle deletion of the special task notifier. */

  std::lock_guard<std::mutex> guard(_task_lock);

  if (this == _notifier)
    _notifier = NULL;
}

/* Handle a deletion event for an object. */

void ObjectNotifier::release(notifiable_t *object)
{
  std::lock_guard<std::mutex> guard(_monitored_lock);

  _monitored.erase(object);
}

/* Subscribe to notification events on an object. */

long ObjectNotifier::subscribe(notifiable_t *object, notify_flags_t flags)
{
  std::lock_guard<std::mutex> guard(_monitored_lock);

  /* Ensure that a handler resource is available for object notifications. */

  NotifierResource *resource;

  if (object->handler != NULL)
    resource = reinterpret_cast<NotifierResource *>(object->handler);

  /* Create a resource if none is recorded. */

  else
  {
    resource = new NotifierResource(object);
    ResourceServer server(resource);
    long err = server.start_thread(&resource->endpoint);

    if (err)
      return err;

    object->handler = resource;
  }

  /* Record the object. */

  _monitored.insert(object);

  /* Record this notifier to get general notifications. */

  return resource->add(this, flags);
}

/* Unsubscribe from notification events on an object. */

long ObjectNotifier::unsubscribe(notifiable_t *object)
{
  std::lock_guard<std::mutex> guard(_monitored_lock);

  if (object->handler == NULL)
    return L4_EOK;

  NotifierResource *resource = reinterpret_cast<NotifierResource *>(object->handler);

  _monitored.erase(object);
  return resource->remove(this);
}

/* Handle a notification event for an object. */

void ObjectNotifier::notify(notifiable_t *object)
{
  /* Enter critical section to access the queue. */

  std::lock_guard<std::mutex> guard(_affected_lock);

  /* Ensure that a queue entry exists for the object. */

  NotifiableObjects::iterator it = _affected.find(object);

  if (it == _affected.end())
  {
    _queued.push_back(object);
    _affected.insert(object);
  }

  /* Notify any waiting caller that at least one notification is available. */

  _condition.notify_one();
}

/* Wait for notification events on objects, returning each object that has a
   notification registered for it. */

long ObjectNotifier::wait(notifiable_t **object)
{
  /* Enter critical section to access the queue. */

  std::unique_lock<std::mutex> guard(_affected_lock);

  while (1)
  {
    /* With pending notifications, return the first object. */

    if (!_affected.empty())
    {
      *object = _queued.front();
      _queued.pop_front();
      _affected.erase(*object);
      break;
    }

    /* Otherwise, wait for notifications. */

    _condition.wait(guard);
  }

  return L4_EOK;
}

/* Wait for notification events on a specific object. */

long ObjectNotifier::wait_object(notifiable_t *object)
{
  if (object->handler == NULL)
    return -L4_EINVAL;

  reinterpret_cast<NotifierResource *>(object->handler)->wait();
  return L4_EOK;
}



/* Virtual destructor required for base class instance reference deletion. */

NotifierResource::~NotifierResource()
{
}

/* Handle the release of the resource. */

void NotifierResource::close()
{
  _release();
  _object->handler = NULL;
}

/* Object-specific resource methods. */

ipc_server_default_config_type NotifierResource::config()
{
  return config_Notifier;
}



/* Registration of notifiers. */

long NotifierResource::add(ObjectNotifier *notifier, notify_flags_t flags)
{
  std::lock_guard<std::mutex> guard(_lock);

  bool is_new = _notifiers.empty();

  _notifiers.insert(notifier);

  /* Subscribe, sending the notification endpoint via the principal reference
     for the object. */

  if (!is_new)
    return L4_EOK;

  client_Notification notify(_object->base->ref);

  return notify.subscribe(endpoint, flags);
}

long NotifierResource::remove(ObjectNotifier *notifier)
{
  std::lock_guard<std::mutex> guard(_lock);

  _notifiers.erase(notifier);

  /* Unsubscribe via the notification interface. */

  if (!_notifiers.empty())
    return L4_EOK;

  client_Notification notify(_object->base->ref);

  return notify.unsubscribe();
}



/* Register a notification received by an object-specific resource. */

long NotifierResource::notify(notify_flags_t flags, notify_values_t values)
{
  /* Update the notifiable object. */

  _object->notifications |= flags;
  _object->values = values;

  _notify();
  return L4_EOK;
}

void NotifierResource::_notify()
{
  /* Enter critical section for the resource. */

  std::lock_guard<std::mutex> guard(_lock);

  /* Record a pending notification which persists if nothing is waiting. */

  _pending = true;

  /* Notify any party waiting specifically on this object. */

  _condition.notify_one();

  /* Register the notification with all notifier objects. */

  ObjectNotifiers::iterator it;

  for (it = _notifiers.begin(); it != _notifiers.end(); it++)
    (*it)->notify(_object);
}

void NotifierResource::_release()
{
  /* Enter critical section for the resource. */

  std::lock_guard<std::mutex> guard(_lock);

  ObjectNotifiers::iterator it;

  for (it = _notifiers.begin(); it != _notifiers.end(); it++)
    (*it)->release(_object);
}

/* Wait for notification events on a specific object. */

long NotifierResource::wait()
{
  /* Enter critical section for the resource. */

  std::unique_lock<std::mutex> guard(_lock);

  /* Wait for the notification condition. */

  if (!_pending)
    _condition.wait(guard);

  _pending = false;
  return L4_EOK;
}

/* vim: tabstop=2 expandtab shiftwidth=2
*/
